use component::{Component, LineArea, Plan, Widget};
use std::{
    io::{Stdout, Write},
    time::Duration,
};
use termion::{
    clear, cursor,
    input::{MouseTerminal, TermRead},
    raw::{IntoRawMode, RawTerminal},
    screen::{AlternateScreen, IntoAlternateScreen},
    AsyncReader,
};
use theme::{Color, ColorSet};
use token::Token;
use view::Event;

type Result<T> = std::result::Result<T, anyhow::Error>;
type Screen = MouseTerminal<RawTerminal<AlternateScreen<Stdout>>>;

extern crate termion;
pub mod component;
pub mod theme;
pub mod token;
pub mod view;

pub async fn run<V>(view: V) -> Result<()>
where
    V: View,
{
    let mut view = view;
    let mut screen: Screen = MouseTerminal::from(
        std::io::stdout().into_alternate_screen()?.into_raw_mode()?,
    );
    let mut events_iter = termion::async_stdin().events();
    let mut plans = view.init()?;
    let (term_w, term_h) = termion::terminal_size()?;
    let mut layers = make_layers(
        plans,
        ColorSet {
            fg: Color::WHITE,
            bg: Color::BLACK,
        },
        (term_w as usize, term_h as usize),
    );
    render_layers(&layers, &mut screen)?;
    screen.flush()?;
    loop {
        let event = match events_iter.next() {
            Some(e) => e,
            None => {
                std::thread::sleep(Duration::from_millis(50));
                continue;
            }
        }?;

        let event = match event {
            termion::event::Event::Mouse(mus) => match mus {
                termion::event::MouseEvent::Press(_, x, y) => {
                    // TODO: scrolling later
                    match match_click(&layers, (x, y)) {
                        Some(yay) => Event::Link(yay),
                        None => continue,
                    }
                }
                _ => Event::Input(event),
            },
            _ => Event::Input(event),
        };

        plans = view.update(event)?;
        let (term_w, term_h) = termion::terminal_size()?;
        layers = make_layers(
            plans,
            ColorSet {
                fg: Color::WHITE,
                bg: Color::BLACK,
            },
            (term_w as usize, term_h as usize),
        );
        write!(screen, "{}", clear::All)?;
        render_layers(&layers, &mut screen)?;
        screen.flush()?;
    }
}

fn match_click(
    layers: &Vec<Layer>,
    click: (u16, u16),
) -> Option<String> {
    for layer in layers.into_iter().rev() {
        for link in &layer.links {
            if link.1.matches(click) {
                return Some(link.0.clone());
            }
        }
    }
    None
}

fn make_layers(
    plans: PlanLayers,
    base_colorset: ColorSet,
    term_dimensions: (usize, usize),
) -> Vec<Layer> {
    plans
        .into_iter()
        .map(|plan| Layer::make(plan, base_colorset, term_dimensions))
        .collect()
}

fn render_layers(
    layered: &Vec<Layer>,
    scr: &mut Screen,
) -> Result<()> {
    for layer in layered.into_iter().map(|layer| {
        let mut line = 1;
        (&layer.components)
            .into_iter()
            .map(|component| match component {
                Component::NextLine => {
                    line += 1;
                    "\n\r".into()
                }
                Component::Link(_, _) => panic!(
                    "links should be filtered out at the layer"
                ),
                Component::X(x) => {
                    cursor::Goto((x + 1) as u16, line).into()
                }
                Component::String(s) => s.clone(),
                Component::Fg(c) => c.fg(),
                Component::Bg(c) => c.bg(),
            })
            .collect::<String>()
    }) {
        write!(scr, "{}", layer)?;
    }
    Ok(())
}

pub struct Layer {
    links: Vec<(String, LineArea)>,
    components: Vec<Component>,
}

impl Layer {
    pub fn make(
        plan: Plan,
        base_colorset: ColorSet,
        term_dimensions: (usize, usize),
    ) -> Self {
        let mut line = 1;
        let mut line_offset = 1;
        let mut links = Vec::new();
        let components = plan
            .make(base_colorset, term_dimensions)
            .into_iter()
            .filter(|comp| match comp {
                Component::NextLine => {
                    line += 1;
                    line_offset = 1;
                    true
                }
                Component::X(x) => {
                    line_offset = (x + 1) as u16;
                    true
                }
                Component::String(s) => {
                    line_offset += s.len() as u16;
                    true
                }
                Component::Link(name, length) => {
                    links.push((
                        name.clone(),
                        LineArea::new(
                            (line_offset, line),
                            *length as u16,
                        ),
                    ));
                    false
                }
                _ => true,
            })
            .collect();
        Self { links, components }
    }
}
// Simple type alias to be clear that the plans are going to
// be layered on top of eachother
pub type PlanLayers = Vec<Plan>;

pub trait View {
    fn init(
        &mut self,
    ) -> std::result::Result<PlanLayers, anyhow::Error>;
    fn update(
        &mut self,
        event: Event,
    ) -> std::result::Result<PlanLayers, anyhow::Error>;
}
